/*
 * matchreply
 *
 * matchreply checks whether a mail message is a reply to any message
 * in the specified Maildir folders.  matchreply is designed to be
 * called from procmail like delivery agents.  Currently, inotify uses
 * inotify system call to watch changes in Maildir folders.  inotify
 * is available from Linux 2.6.13.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * Copyright (C) Tejun Heo <htejun@gmail.com>, 2005
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <dirent.h>
#include <ctype.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <signal.h>
#include <inttypes.h>
#include <sys/poll.h>
#include <time.h>
#include <stdarg.h>
#include <sys/inotify.h>

/* Why isn't this defined in sys/un.h? */
#ifndef UNIX_PATH_MAX
#define UNIX_PATH_MAX		108
#endif

/*
 * As of this writing, most distributions still don't ship
 * inotify-aware glibc.  Use proper -D*_DEFINE_INOTIFY_SYSCALLS
 * compile option in such cases.
 */
#if defined(X86_64_DEFINE_INOTIFY_SYSCALLS)

#include <asm/unistd.h>
#define __NR_inotify_init	253
#define __NR_inotify_add_watch	254
#define __NR_inotify_rm_watch	255

_syscall0(int, inotify_init)
_syscall3(int, inotify_add_watch, int, fd, const char *, path, uint32_t, mask)
_syscall2(int, inotify_rm_watch, int, fd, int32_t, wd)

#elif defined(I386_DEFINE_INOTIFY_SYSCALLS)

#include <asm/unistd.h>
#define __NR_inotify_init       291
#define __NR_inotify_add_watch  292
#define __NR_inotify_rm_watch   293

_syscall0(int, inotify_init)
_syscall3(int, inotify_add_watch, int, fd, const char *, path, uint32_t, mask)
_syscall2(int, inotify_rm_watch, int, fd, int32_t, wd)

#endif

static const char *helpmsg =
"\n"
"Usage: matchreply [OPTION] -i IDSTR MAILDIRS...\n"
"\n"
"  -v             verbose\n"
"  -d             match duplicates\n"
"  -l LOGFILE     file to append logs to\n"
"  -H HASHSIZE    hashtable size, default is 65536\n"
"  -V             print version and exit\n"
"  -h             print this help message and exit\n"
"\n";

#define MATCHREPLY_VERSION	"0.11"
#define MATCHREPLY_HELLO	"Hi there! I'm matchreply "MATCHREPLY_VERSION

/*
 * Hashtable lookup doesn't occur very often (mostly only on mail
 * deliveries).  65536 should be large enough for most cases.  This
 * will consume 1M on 64bit machines.
 */
#define DFL_MSGID_HASHSIZE	65536
/*
 * We assume In-reply-to: and References: appear in the first
 * MAIL_HEADER_MAX_SIZE bytes of a mail message.
 */
#define MAIL_HEADER_MAX_SIZE	32768
#define MAIL_MAX_REFS		128

#define INOTIFY_MASK		(IN_MOVE|IN_CREATE|IN_DELETE|IN_DELETE_SELF)

static const char msgid_header[] = "Message-id:";
static const char inreplyto_header[] = "In-reply-to:";
static const char references_header[] = "References:";

struct msgid {
	struct msgid *id_next, *file_next;
	char buf[]; /* id'\0'file'\0' */
};

static char socket_file[UNIX_PATH_MAX];
static unsigned long conf_hash;
static int verbose;
FILE *log_fp;
static char *instance_id, log_id[64];

static const char **mfolders;
static int nr_mfolders, inotify_fd;
static int *inotify_wds;

static struct msgid **msgid_hashtbl, **file_hashtbl;
static int match_duplicate;
static int hashtbl_len = DFL_MSGID_HASHSIZE, nr_msgids;
static int nr_cols_msgid, nr_dcols_msgid, nr_cols_file, nr_dcols_file;

#define log_info(fmt, args...) \
	log_message("INF", fmt , ##args)

#define log_very_verbose(fmt, args...) do { \
	if (verbose > 1) log_message("DBG", fmt , ##args); \
	} while (0)

#define log_verbose(fmt, args...) do { \
	if (verbose) log_message("DBG", fmt , ##args); \
	} while (0)

#define log_error(fmt, args...) \
	log_message("ERR", fmt , ##args)

#define log_errno(fmt, args...) \
	log_message("ERR", fmt" (%s)" , ##args, strerror(errno))

static void log_message(const char *type, const char *fmt, ...)
{
	time_t t;
	struct tm tm;
	char buf[1024];
	int len, n;
	va_list ap;

	len = sizeof(buf) - 2;
	buf[0] = '[';
	n = 1;

	time(&t);
	localtime_r(&t, &tm);
	n += strftime(buf + n, len - n, "%F %T", &tm);

	n += snprintf(buf + n, len - n, " %s %s] ", log_id, type);
	if (n > len)
		n = len;

	va_start(ap, fmt);
	n += vsnprintf(buf + n, len - n, fmt, ap);
	if (n > len)
		n = len;
	va_end(ap);

	buf[n] = '\n';
	buf[n + 1] = '\0';
	fputs(buf, log_fp);
}

static void install_term_sighandlers(void (*handler)(int))
{
	signal(SIGHUP, handler);
	signal(SIGINT, handler);
	signal(SIGQUIT, handler);
	signal(SIGTERM, handler);
}

static const char * find_header(const char *header, int header_len,
				const char *buf, size_t buflen, int *plen)
{
	const char *cp, *rend, *end, *eol;

	if (buflen <= header_len)
		return NULL;

	cp = buf;
	rend = buf + buflen;
	end = rend - header_len;

	while (cp < end) {
		if (*cp == '\n' || (cp[0] == '\r' && cp[1] == '\n'))
			return NULL;
		if (strncasecmp(cp, header, header_len) == 0)
			break;
		while (cp < end && *cp != '\n')
			cp++;
		cp++;
	}

	cp += header_len;
	eol = cp;
	while (eol < rend) {
		if (*eol == '\n' && (eol == rend - 1 || !isspace(eol[1])))
			break;
		eol++;
	}

	if (cp < eol) {
		*plen = eol - cp;
		return cp;
	}
	return NULL;
}

static inline int parse_ids(const char *buf, int buf_len,
			    int max_ids, const char **ids, int *id_lens)
{
	int i;
	const char *p = buf, *end = buf + buf_len;
	const char *id;

	for (i = 0; i < max_ids; i++) {
		while (p < end && isspace(*p))
			p++;
		if (p == end || *p != '<')
			break;
		p++;
		id = p;
		while (p < end && *p != '>')
			p++;
		if (p == end || p == id)
			break;
		*(ids++) = id;
		*(id_lens++) = p - id;
		p++;
	}

	return i;
}

static inline unsigned long calc_hash(const char *p, int len)
{
        unsigned long hash = 5381;

        while (len--) {
		hash ^= (hash << 4) + *p;
		p++;
	}

        return hash;
}

static inline void parse_maildir_filename(const char *file, int *flen,
					  int *trashed)
{
	const char *p = file + *flen - 1;

	*trashed = 0;
	while (file < p && *p != ':') {
		*trashed |= *p == 'T';
		p--;
	}
	if (file < p)
		*flen = p - file;
}

static inline int make_msgid(const char *id, int idlen,
			     const char *file, int uniqlen)
{
	struct msgid *msgid;
	unsigned idx;

	msgid = malloc(sizeof(*msgid) + idlen + uniqlen + 2);
	if (msgid == NULL)
		return -1;

	memcpy(msgid->buf, id, idlen);
	msgid->buf[idlen] = '\0';
	memcpy(msgid->buf + idlen + 1, file, uniqlen);
	msgid->buf[idlen + 1 + uniqlen] = '\0';

	idx = calc_hash(id, idlen) % hashtbl_len;
	msgid->id_next = msgid_hashtbl[idx];
	msgid_hashtbl[idx] = msgid;
	if (msgid->id_next) {
		nr_cols_msgid++;
		nr_dcols_msgid += msgid->id_next->id_next != NULL;
	}

	idx = calc_hash(file, uniqlen) % hashtbl_len;
	msgid->file_next = file_hashtbl[idx];
	file_hashtbl[idx] = msgid;
	if (msgid->file_next) {
		nr_cols_file++;
		nr_dcols_file += msgid->file_next->file_next != NULL;
	}

	if (++nr_msgids > hashtbl_len * 5) {
		static int printed_warning = 0;
		if (!printed_warning) {
			printed_warning = 1;
			log_error("Gee... nr_msgids exceeded 5 times of "
				  "hashtbl_len, consider enlarging hashtbl_len");
		}
	}

	log_very_verbose("inserted %d %d %d %d %d <%s> '%s'",
			 nr_msgids, nr_cols_msgid, nr_dcols_msgid,
			 nr_cols_file, nr_dcols_file,
			 msgid->buf, msgid->buf + idlen + 1);

	return 0;
}

static int maildir_insert_one(const char *dir, const char *file)
{
	char path[PATH_MAX];
	int dlen, flen, flen_uniq, trashed, id_len, fd;
	struct stat sb;
	const char *id;
	void *map;
	int ret = -1;

	dlen = strlen(dir);
	flen = strlen(file);

	flen_uniq = flen;
	parse_maildir_filename(file, &flen_uniq, &trashed);

	if (trashed) {
		log_verbose("skipping trashed mail '%s'", file);
		return -1;
	}

	if (dir[dlen - 1] == '/')
		dlen--;

	if (dlen + flen + 1 >= PATH_MAX) {
		log_error("dir='%s' file='%s' too long", dir, file);
		return -1;
	}

	memcpy(path, dir, dlen);
	path[dlen] = '/';
	memcpy(path + dlen + 1, file, flen + 1);

	if ((fd = open(path, O_RDONLY)) < 0) {
		/*
		 * If the file doesn't exist, it probably means that
		 * another mail reader already has moved the file.
		 * update_hashtbls will take care of it.
		 */
		if (errno == ENOENT)
			log_verbose("'%s' non-existent, already moved?", path);
		else
			log_errno("failed to open '%s'", path);
		return -1;
	}

	if (fstat(fd, &sb) < 0) {
		log_errno("failed to stat '%s'", path);
		goto out_close;
	}

	map = mmap(NULL, sb.st_size, PROT_READ, MAP_SHARED, fd, 0);
	if (map == MAP_FAILED) {
		log_errno("failed to map '%s'", path);
		goto out_close;
	}

	/*
	 * In most cases, headers will be in the first page.  Don't
	 * flush page cache pages by unnecessary readahead.
	 */
	madvise(map, sb.st_size, MADV_RANDOM);

	id = find_header(msgid_header, sizeof(msgid_header) - 1,
			 map, sb.st_size, &id_len);
	if (id == NULL) {
		log_verbose("No Message-id in '%s'", path);
		goto out_unmap;
	}

	if (parse_ids(id, id_len, 1, &id, &id_len) != 1) {
		log_error("Malformed Message-id in '%s'", path);
		goto out_unmap;
	}

	if (make_msgid(id, id_len, file, flen_uniq) < 0) {
		log_errno("insert_msgid");
		goto out_unmap;
	}

	ret = 0;

 out_unmap:
	munmap(map, sb.st_size);
 out_close:
	close(fd);

	return ret;
}

static struct msgid ** lookup_msgidp_file(const char *file, int uniqlen,
					  struct msgid **headp)
{
	struct msgid **pp, *p;
	char *cp;

	pp = file_hashtbl + (calc_hash(file, uniqlen) % hashtbl_len);

	if (headp)
		*headp = *pp;

	while (*pp) {
		p = *pp;
		cp = p->buf;

		while (*cp != '\0')
			cp++;
		cp++;

		if (strlen(cp) == uniqlen && memcmp(cp, file, uniqlen) == 0)
			return pp;

		pp = &p->file_next;
	}

	return NULL;
}

static int maildir_remove_one(const char *file, int uniqlen)
{
	struct msgid **pp, *p, *head;

	pp = lookup_msgidp_file(file, uniqlen, &head);

	if (!pp) {
		log_info("removal failed, msgid for '%s' doesn't exist", file);
		return 0;
	}

	/* remove from file hash */
	p = *pp;
	*pp = p->file_next;
	if (head) {
		nr_cols_file--;
		nr_dcols_file -= head->file_next != NULL;
	}

	/* remove from msgid hash */
	pp = msgid_hashtbl + (calc_hash(p->buf, strlen(p->buf)) % hashtbl_len);
	head = *pp;

	while (*pp && *pp != p)
		pp = &(*pp)->id_next;

	if (!*pp) {
		log_error("<%s> '%s' is linked on fhash but not on ihash",
			  p->buf, file);
		return -1;
	}

	*pp = p->id_next;
	if (head) {
		nr_cols_msgid--;
		nr_dcols_msgid -= head->id_next != NULL;
	}

	nr_msgids--;

	log_very_verbose("removed %d %d %d %d %d <%s> '%s'",
			 nr_msgids, nr_cols_msgid, nr_dcols_msgid,
			 nr_cols_file, nr_dcols_file, p->buf, file);
	free(p);
	return 0;
}

static int build_idx(int nr_targets, const char **targets)
{
	int i;

	for (i = 0; i < nr_targets; i++) {
		const char *targ = targets[i];
		struct stat sb;

		if (stat(targ, &sb) < 0) {
			log_errno("failed to stat '%s'", targ);
			continue;
		}

		if (S_ISDIR(sb.st_mode)) {
			/* assume Maildir format */
			DIR *dir;
			struct dirent *dirent;
			int wd;

			if ((dir = opendir(targ)) == NULL) {
				log_errno("failed to opendir '%s'", targ);
				continue;
			}

			if (inotify_fd >= 0) {
				wd = inotify_add_watch(inotify_fd, targ,
						       INOTIFY_MASK);
				if (wd >= 0) {
					inotify_wds[i] = wd;
				} else {
					log_errno("failed to add watch to '%s', "
						  "reverting to oneshot mode",
						  targ);
					inotify_fd = -1;
				}
			}

			while ((dirent = readdir(dir)) != NULL) {
				char *name = dirent->d_name;
				if (name[0] == '.')
					continue;
				maildir_insert_one(targ, name);
			}
			closedir(dir);
		} else {
			/* maybe support mbox format someday? */
			log_error("%s is not a directory", targ);
			return -1;
		}
	}

	log_info("hashtbls built, len=%d nr_msgids=%d "
		 "MSGID_HASH cols=%d dcols=%d FILE_HASH cols=%d dcols=%d",
		 hashtbl_len, nr_msgids,
		 nr_cols_msgid, nr_dcols_msgid, nr_cols_file, nr_dcols_file);

	return 0;
}

static inline struct inotify_event *
get_next_ie(struct inotify_event *ie, int *left)
{
	int len;

	len = sizeof(*ie) + ie->len;
	*left -= len;
	return *left ? (void *)ie + len : NULL;
}

static inline int update_hashtbls(int first)
{
	char buf[4096];
	int ret, left;
	struct inotify_event *ie, *next_ie;

	if (inotify_fd < 0)
		return 0;

	ret = read(inotify_fd, buf, sizeof(buf));
	if (ret <= 0) {
		if (ret == 0) {
			log_error("EOF from inotify_fd");
			return -1;
		} else if (errno != EAGAIN) {
			log_errno("failed to read from inotify");
			return -1;
		} else
			return 0;
	}

	ie = (void *)buf;
	left = ret;
	do {
		int i, uniqlen, next_uniqlen, trashed, next_trashed;
		int op = 0;	/* 0 == noop, 1 == insert, 2 == remove */

		next_ie = get_next_ie(ie, &left);

		/* check for error conditions */
		if (ie->mask & (IN_UNMOUNT | IN_IGNORED | IN_DELETE_SELF)) {
			log_error("something happened to Maildir, bailing out");
			return -1;
		}

		if (ie->mask & IN_Q_OVERFLOW) {
			if (first) {
				log_error("inotify queue overflow on the first "
					 "service, reverting to oneshot mode");
				inotify_fd = -1;
			} else {
				log_error("inotify queue overflow, bailing out");
				return -1;
			}
		}

		if (!ie->len) {
			log_error("ie->len == 0 after handling exception "
				  "events? (mask=%x)", ie->mask);
			continue;
		}

		/* parse filename.  note that ie->len != strlen */
		uniqlen = strlen(ie->name);
		parse_maildir_filename(ie->name, &uniqlen, &trashed);

		if ((ie->mask & IN_MOVED_FROM) && next_ie &&
		    (next_ie->mask & IN_MOVED_TO) &&
		    ie->cookie == next_ie->cookie) {
			if (!next_ie->len) {
				log_error("next_ie->len == 0 on MOVED_TO?");
				continue;
			}

			/* parse filename.  note that ie->len != strlen */
			next_uniqlen = strlen(next_ie->name);
			parse_maildir_filename(next_ie->name, &next_uniqlen,
					       &next_trashed);

			/* new/ -> cur/ move or flag change */
			switch (trashed | (next_trashed << 1)) {
			case 0:
				if (lookup_msgidp_file(ie->name, uniqlen, NULL))
					goto rename_noop;
				/* We missed this file previously, insert now */
			case 1:
				log_verbose("RENAME insert: '%s' -> '%s'",
					    ie->name, next_ie->name);
				op = 1;
				break;
			case 2:
				log_verbose("RENAME remove: '%s' -> '%s'",
					    ie->name, next_ie->name);
				op = 2;
				break;
			case 3:
			rename_noop:
				log_verbose("RENAME noop: '%s' -> '%s'",
					    ie->name, next_ie->name);
				break;
			}
			ie = next_ie;
		} else if (ie->mask & (IN_MOVED_TO | IN_CREATE)) {
			if (!trashed) {
				log_verbose("CREATE insert: '%s'", ie->name);
				op = 1;
			} else
				log_verbose("CREATE noop: '%s'", ie->name);
		} else if (ie->mask & (IN_MOVED_FROM | IN_DELETE)) {
			if (!trashed) {
				log_verbose("DELETE remove: '%s'", ie->name);
				op = 2;
			} else
				log_verbose("DELETE noop: '%s'", ie->name);
		}

		switch (op) {
		case 0:
			break;
		case 1:
			/* look for matching mfolder */
			for (i = 0; i < nr_mfolders; i++)
				if (ie->wd == inotify_wds[i])
					break;
			if (i == nr_mfolders) {
				log_error("invalid wd reported");
				return -1;
			}
			/* insert */
			log_verbose("new mail: '%s'", ie->name);
			maildir_insert_one(mfolders[i], ie->name);
			break;
		case 2:
			log_verbose("removing mail: '%s'", ie->name);
			if (maildir_remove_one(ie->name, uniqlen) < 0)
				return -1;
			break;
		}

		if (ie == next_ie)
			next_ie = get_next_ie(ie, &left);
	} while ((ie = next_ie) != NULL);

	return 0;
}

static inline int match_id(const char *id, int id_len)
{
	struct msgid *msgid;

	msgid = msgid_hashtbl[calc_hash(id, id_len) % hashtbl_len];
	while (msgid) {
		if (strcmp(msgid->buf, id) == 0)
			return 1;
		msgid = msgid->id_next;
	}
	return 0;
}

static void server_atexit(void)
{
	unlink(socket_file);
}

static void server_termsig_handler(int sig)
{
	exit(1);
}

static int server_main(int sockfd)
{
	size_t sz;
	unsigned long clt_conf_hash;
	struct pollfd pfds[2];
	int nfds, first = 1;

	snprintf(log_id, sizeof(log_id) - 1,
		 "%s SVR %d", instance_id, getpid());
	log_info("starting up...");

	/*
	 * As the client which forked us will soon exit, the following
	 * is enough for daemonization.
	 */
	setsid();

	/*
	 * On entry to this function, all term signals are set to
	 * SIG_IGN.  Install exit handler and term sig handlers to
	 * remove socket file on exit.
	 */
	atexit(server_atexit);
	install_term_sighandlers(server_termsig_handler);

	/* allocate hash tables & inotify wd array */
	sz = sizeof(msgid_hashtbl[0]) * hashtbl_len;
	msgid_hashtbl = malloc(sz);
	file_hashtbl = malloc(sz);
	if (msgid_hashtbl == NULL || file_hashtbl == NULL) {
		log_errno("failed to allocate hashtbl");
		return 1;
	}
	memset(msgid_hashtbl, 0, sz);
	memset(file_hashtbl, 0, sz);

	sz = sizeof(inotify_wds[0]) * hashtbl_len;
	inotify_wds = malloc(sz);
	if (inotify_wds == NULL) {
		log_errno("failed to allocate inotify_wds");
		return 1;
	}
	memset(inotify_wds, -1, sz);

	/* init inotify & build idx */
	inotify_fd = inotify_init();

	if (build_idx(nr_mfolders, mfolders) < 0)
		return 1;

	memset(pfds, 0, sizeof(pfds));

	fcntl(sockfd, F_SETFL, O_NONBLOCK);
	nfds = 1;
	pfds[0].fd = sockfd;
	pfds[0].events = POLLIN;

	if (inotify_fd >= 0) {
		fcntl(inotify_fd, F_SETFL, O_NONBLOCK);
		nfds++;
		pfds[1].fd = inotify_fd;
		pfds[1].events = POLLIN;
	}

	while (1) {
		int fd, idx, match, ret;
		FILE *fp;
		char buf[512];

		if (poll(pfds, nfds, -1) < 0) {
			log_errno("poll");
			break;
		}

		if (update_hashtbls(first) < 0)
			break;

		fd = accept(sockfd, NULL, NULL);
		if (fd < 0) {
			if (errno == EAGAIN || errno == EWOULDBLOCK)
				continue;
			log_errno("accept");
			break;
		}
		first = 0;

		log_verbose("connection accepted");

		if ((fp = fdopen(fd, "a+")) == NULL) {
			log_errno("fdopen");
			close(fd);
			continue;
		}

		if (fgets(buf, sizeof(buf), fp) == NULL) {
			log_error("failed to read HELLO");
			goto next;
		}

		ret = sscanf(buf, MATCHREPLY_HELLO" %lu",
			     &clt_conf_hash);
		if (ret != 1) {
			log_error("HELLO mismatch, bailing out");
			break;
		}

		if (clt_conf_hash != conf_hash) {
			log_error("configuration mismatch, bailing out");
			break;
		}

		idx = 0;
		match = -1;
		while ((fgets(buf, sizeof(buf), fp)) != NULL) {
			size_t len;
			len = strlen(buf);
			if (len <= 1)
				break;
			buf[--len] = '\0';
			if (match < 0) {
				if (match_id(buf, len)) {
					log_verbose("%02d: <%s> hit", idx, buf);
					match = idx;
				} else
					log_verbose("%02d: <%s> miss", idx, buf);
			}
			idx++;
		}
		fprintf(fp, "%s %d\n", MATCHREPLY_HELLO, match);
	next:
		fclose(fp);
		log_verbose("service complete");
		if (inotify_fd < 0) {
			log_info("oneshot mode, exiting...");
			break;
		}
	}

	return 0;
}

static int start_server(void)
{
	int sock, ret = 0;
	struct sockaddr_un sun;
	pid_t pid;

	log_info("trying to start a server...");

	if ((sock = socket(PF_UNIX, SOCK_STREAM, 0)) < 0) {
		log_errno("socket");
		return -1;
	}

	/*
	 * Plug term signals until server is started to prevent
	 * dangling socket file.  Server will install its own handlers
	 * to remove socket file on exit.
	 */
	install_term_sighandlers(SIG_IGN);

	memset(&sun, 0, sizeof(sun));
	sun.sun_family = AF_UNIX;
	strncpy(sun.sun_path, socket_file, UNIX_PATH_MAX - 1);

	if (bind(sock, (struct sockaddr *)&sun, sizeof(sun)) < 0) {
		if (errno != EADDRINUSE) {
			log_errno("failed to bind server socket to '%s'",
				  socket_file);
			ret = -1;
		} else
			log_info("EADDRINUSE, somebody started server already?");
		goto out;
	}

	chmod(socket_file, 0700);	/* XXX not portable XXX */

	if (listen(sock, 5) < 0) {
		if (errno != EADDRINUSE) {
			log_errno("listen");
			ret = -1;
		} else
			log_info("EADDRINUSE, somebody started server already?");
		goto out_unlink;
	}

	pid = fork();

	if (pid > 0 || pid < 0) {
		/* in client */
		if (pid < 0) {
			log_errno("fork");
			ret = -1;
			goto out_unlink;
		}
	} else if (pid == 0) {
		/* in server process */
		exit(server_main(sock));
	}
	log_info("successfully started server [pid=%d]", pid);
 out:
	close(sock);
	install_term_sighandlers(SIG_DFL);
	return ret;
 out_unlink:
	unlink(socket_file);
	goto out;
}

static int establish_conn(void)
{
	int sock;
	struct sockaddr_un sun;
	int retry_cnt = 0;

	if ((sock = socket(PF_UNIX, SOCK_STREAM, 0)) < 0) {
		log_errno("socket");
		return -1;
	}

	memset(&sun, 0, sizeof(sun));
	sun.sun_family = AF_UNIX;
	strncpy(sun.sun_path, socket_file, UNIX_PATH_MAX - 1);

 retry:
	if (connect(sock, (struct sockaddr *)&sun, sizeof(sun)) == 0)
		return sock;

	switch (errno) {
	case ECONNREFUSED:
		/* might be starting up, retry */
		if (retry_cnt++ < 5) {
			sleep(1);
			goto retry;
		}
		unlink(socket_file);
		errno = ENOENT;
		/* fall through */
	case ENOENT:
		/* server not present */
		close(sock);
		return -1;
	default:
		log_errno("failed to connect to '%s'", socket_file);
		close(sock);
		return -1;
	}
}

static int client_main(void)
{
	int retry_cnt = 5;
	const char *self_id, *ids[MAIL_MAX_REFS];
	int self_id_len, id_lens[MAIL_MAX_REFS];
	const char *header;
	char *buf;
	int buf_len, i, ret, nr, sock, header_len;
	FILE *fp;

	/* Let's ignore SIGPIPE */
	signal(SIGPIPE, SIG_IGN);

	/* Read in mail body */
	buf_len = MAIL_HEADER_MAX_SIZE;
	buf = malloc(buf_len);
	ret = fread(buf, 1, buf_len, stdin);
	if (ret == 0) {
		log_error("failed to read from stdin");
		return 1;
	}

	/* Parse message-id for logging */
	header = find_header(msgid_header, sizeof(msgid_header) - 1,
			     buf, ret, &header_len);
	self_id = "< NO_ID >";
	if (header && parse_ids(header, header_len, 1, &self_id, &self_id_len))
		((char *)self_id)[self_id_len] = '\0';

	/* Parse in-reply-to and references */
	nr = 0;
	header = find_header(inreplyto_header, sizeof(inreplyto_header) - 1,
			     buf, ret, &header_len);
	if (header)
		nr += parse_ids(header, header_len,
				MAIL_MAX_REFS - nr, ids + nr, id_lens + nr);

	header = find_header(references_header, sizeof(references_header) - 1,
			     buf, ret, &header_len);
	if (header)
		nr += parse_ids(header, header_len,
				MAIL_MAX_REFS - nr, ids + nr, id_lens + nr);

	if (nr == 0 && !match_duplicate) {
		log_info("message <%s> has no reference", self_id);
		return 1;
	}

 retry:
	/* Establish connection.  If server is not there, start it. */
	while ((sock = establish_conn()) < 0 && errno == ENOENT) {
		/* We don't have server.  Start one */
		if (start_server() < 0)
			return 1;
	}

	if (sock < 0) {
		log_error("failed to connect to server");
		return 1;
	}

	if ((fp = fdopen(sock, "a+")) == NULL) {
		log_errno("fdopen");
		return 1;
	}

	fprintf(fp, "%s %lu\n", MATCHREPLY_HELLO, conf_hash);
	fprintf(fp, "%s\n", self_id);
	for (i = 0; i < nr; i++) {
		((char *)(ids[i]))[id_lens[i]] = '\0';
		fprintf(fp, "%s\n", ids[i]);
	}
	fprintf(fp, "\n");

	/* Read response */
	if (fscanf(fp, MATCHREPLY_HELLO" %d", &ret) != 1) {
		if (retry_cnt--) {
			/*
			 * Maybe just configuration mismatch?
			 */
			fclose(fp);
			goto retry;
		}
		log_error("giving up on protocol mismatch");
		return 1;
	}

	/* Be nice and drain stdin to save our feeder from EPIPE */
	while (fread(buf, 1, buf_len, stdin))
		;

	log_info("<%s>: %d refs, match %sfound idx=%d",
		 self_id, nr, ret >= 0 ? "" : "not ", ret);

	if (match_duplicate)
		return !(ret == 0);
	return !(ret > 0);
}

int main(int argc, char **argv)
{
	const char *logfile = NULL;
	char *endp;
	int c, i, ret;

	while ((c = getopt(argc, argv, "i:dvl:H:S:Vh")) >= 0) {
		switch (c) {
		case 'i':
			instance_id = optarg;
			break;
		case 'd':
			match_duplicate = 1;
			break;
		case 'v':
			verbose++;
			break;
		case 'l':
			logfile = optarg;
			break;
		case 'H':
			hashtbl_len = strtoul(optarg, &endp, 0);
			if (optarg == endp || *endp != '\0' ||
			    hashtbl_len <= 0) {
				fprintf(stderr, "invalid hashsize '%s'\n",
					optarg);
				return 1;
			}
			break;
		case 'V':
			fprintf(stderr, "matchreply v%s\n", MATCHREPLY_VERSION);
			return 1;
		case 'h':
			/* fall through */
		default:
			goto help;
		}
	}

	if (instance_id == NULL || optind == argc) {
	help:
		fprintf(stderr, "%s", helpmsg);
		return 1;
	}

	ret = snprintf(socket_file, UNIX_PATH_MAX, "/tmp/.matchreply-%u-%s",
		       geteuid(), instance_id);
	if (ret >= UNIX_PATH_MAX) {
		log_error("socket_file path '%s' too long", socket_file);
		return 1;
	}

	mfolders = (const char **)argv + optind;
	nr_mfolders = argc - optind;

	/*
	 * Setup loggin
	 */
	snprintf(log_id, sizeof(log_id) - 1,
		 "%s CLT %d", instance_id, getpid());
	log_fp = stderr;
	if (logfile) {
		if ((log_fp = fopen(logfile, "a")) == NULL) {
			fprintf(stderr,
				"failed to open logfile '%s'\n", logfile);
			return 1;
		}
		setlinebuf(log_fp);
	}
	

	/*
	 * Construct configuration hash.  Maybe convert to SHA1
	 * someday.
	 */
	for (i = 0; i < nr_mfolders; i++)
		conf_hash += calc_hash(mfolders[i], strlen(mfolders[i]));
	if (logfile)
		conf_hash += calc_hash(logfile, strlen(logfile));
	conf_hash += hashtbl_len + (verbose << 30);

	return client_main();
}
